#!/usr/bin/env python

##@file GFF3Maker.py

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import glob
import sys
import shutil
from pyRepetUnit.commons.utils.RepetOptionParser import RepetOptionParser
from pyRepetUnit.commons.utils.FileUtils import FileUtils
from pyRepetUnit.commons.sql.DbFactory import DbFactory
from pyRepetUnit.commons.seq.BioseqDB import BioseqDB

## GFF3Maker exports annotations from a 'path' table into a GFF3 file.
#
class GFF3Maker(object):

    def __init__(self, inFastaName = "", patternGFF3 = "", tablesFileName = "", isChado = False, configFileName = "", verbose = 0):
        self._inFastaName = inFastaName
        self._patternGFF3 = patternGFF3
        self._tablesFileName = tablesFileName
        self._isChado = isChado
        self._configFileName = configFileName
        self._iDB = None
        self._verbose = verbose
    
    def setAttributesFromCmdLine(self):
        description = "GFF3Maker with -f option <=> step 1 : create gff files (write only the sequence and not the annotation. Only one sequence in each file)\n"
        description += "GFF3Maker with -g and -t options <=> step 2 : add annotations in each file\n"
        parser = RepetOptionParser(description = description)
        parser.add_option("-f", "--inseq",       dest = "inFastaName",      action = "store",       type = "string", help = "'fasta' file or 'seq' table recording the input sequences (required to generate new '.gff3' files)", default = "")
        parser.add_option("-g", "--gff3pattern", dest = "patternGFF3",      action = "store",       type = "string", help = "gff3 file or pattern (e.g. '*.gff3')", default = "")
        parser.add_option("-t", "--tablesfile",  dest = "tablesFileName",   action = "store",       type = "string", help = "tabulated file of table name to use to create the gff3 files (fields: tier name, format, table name)", default = "")
        parser.add_option("-c", "--chado",       dest = "isChado",          action = "store_true",                   help = "Chado compliance", default = False)
        parser.add_option("-C", "--config",      dest = "configFileName",   action = "store",       type = "string", help = "configuration file for database connection", default = "")
        parser.add_option("-v", "--verbose",     dest = "verbose",          action = "store",       type = "int",    help = "verbosity level (default=0, else 1 or 2)", default = 0)
        (options, args) = parser.parse_args()
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.setInFastaName(options.inFastaName)
        self.setPatternGFF3(options.patternGFF3)
        self.setTablesFileName(options.tablesFileName)
        self.setIsChado(options.isChado)
        self.setConfigFileName(options.configFileName)
        self.setVerbose(options.verbose)
        
    def setInFastaName(self, inFastaName):
        self._inFastaName = inFastaName
        
    def setPatternGFF3(self, patternGFF3):
        self._patternGFF3 = patternGFF3
        
    def setTablesFileName(self, tablesFileName):
        self._tablesFileName = tablesFileName
        
    def setIsChado(self, isChado):
        self._isChado = isChado
        
    def setConfigFileName(self, configFileName):
        self._configFileName = configFileName
        
    def setVerbose(self, verbose):
        self._verbose = verbose

    def checkOptions(self):       
        if self._inFastaName == "" and self._patternGFF3 == "":
            raise Exception("ERROR: options -f or -g required")
    
        if self._configFileName != "":
            if not FileUtils.isRessourceExists(self._configFileName):
                raise Exception("ERROR: configuration file does not exist!")
            
    ## Create as many GFF3 files as sequences given in parameter.
    #
    def _createGff3Files(self):
        if os.path.exists(self._inFastaName):
            if self._verbose > 0:
                print "create GFF3 file(s) from file '%s'..." % self._inFastaName
                sys.stdout.flush()
            inDB = BioseqDB(self._inFastaName)
            for bs in inDB.db:
                outFile = open("%s.gff3" % bs.header, "w")
                outFile.write("##gff-version 3\n")
                outFile.write("##sequence-region %s 1 %s\n" % (bs.header, bs.getLength()))
                outFile.write("##FASTA\n")
                outFile.write(">%s\n" % bs.header)
                i = 0
                while i < bs.getLength():
                    outFile.write(bs.sequence[i:i+60] + "\n")
                    i += 60
                outFile.close()
    
        elif self._iDB.doesTableExist(self._inFastaName):
            if self._verbose > 0:
                print "create GFF3 file(s) from table '%s'..." % self._inFastaName
                sys.stdout.flush()
            qry = "SELECT DISTINCT accession,length,sequence FROM %s" % self._inFastaName
            self._iDB.execute(qry)
            lSequences = self._iDB.fetchall()
            for seq in lSequences:
                outFile = open("%s.gff3" % seq[0], "w")
                outFile.write("##gff-version 3\n")
                outFile.write("##sequence-region %s 1 %s\n" % (seq[0], seq[1]))
                outFile.write("##FASTA\n")
                outFile.write(">%s\n" % seq[0])
                i = 0
                while i < int(seq[1]):
                    outFile.write(seq[2][i:i+60] + "\n")
                    i += 60
                outFile.close()
    
        else:
            print "ERROR: no file or table '%s' exists" % self._inFastaName
            sys.exit(1)
    
    ## Save the annotations from the table corresponding to the given sequence.
    #
    # @param key string sequence name
    # @param table string path table name
    # @param tierName string name of the tier in the GFF3 file
    #
    def _saveAnnotPath(self, key, table, tierName):
        if self._verbose > 1:
            print "processing '%s' for tier '%s'..." % (key, tierName)
            sys.stdout.flush()
    
        shutil.move("%s.gff3" % key, "%s.gff3_tmp" % key)
        tmpFile = open("%s.gff3_tmp" % key, "r")
        line = tmpFile.readline()
        outFile = open("%s.gff3" % key, "w")
        while True:
            if line == "":
                break
            if line != "##FASTA\n":
                outFile.write(line)
                line = tmpFile.readline()
            else:
                source = tierName
                feature = "region"
                frame = "."
                pathString = self._getPathFeatures(table, key, source, feature, frame)
                outFile.write(pathString)
                break
        if line == "##FASTA\n":
            while True:
                if line == "":
                    break
                outFile.write(line)
                line = tmpFile.readline()
        tmpFile.close()
        outFile.close()
        os.remove("%s.gff3_tmp" % key)
    
    ## Retrieve the features to write in the GFF3 file.
    #
    # @param pathTable string name of the table recording the annotations (i.e. the features)
    # @param seqName string name of the sequence (the source feature) on which we want to visualize the matches (the features)
    # @param source string the program that generated the feature (i.e. REPET)
    # @param frame string "." by default (or "+", "-")
    # @return pathString string which will be printed in path file 
    #
    def _getPathFeatures(self, pathTable, seqName, source, feature, frame):
        pathString = ""
        # retrieve all the data about the matches
        qry = "SELECT DISTINCT path, query_start, query_end, subject_name, subject_start, subject_end, E_value, identity FROM %s WHERE query_name=\"%s\"" % (pathTable, seqName)
        self._iDB.execute(qry)
        data = self._iDB.fetchall()
        # organise them into 'match' and 'match_part'
        dPathID2Data = self._gatherSamePathFeatures(data)
        # build the output string
        for pathID in dPathID2Data:
            pathString += self._organizeEachPathFeature(pathID, dPathID2Data[pathID], seqName, source, frame)
        return pathString
    
    ## Gather matches with the same path ID.
    #
    # @param data list of string lists results of a SQL request
    # @return dPathID2Matchs dict whose keys are path IDs and values are matches data
    #
    def _gatherSamePathFeatures(self, data):
        dPathID2Matchs = {}
    
        for i in data:
            pathID = i[0]
            query_start = i[1]
            query_end = i[2]
            subject_name = i[3].replace(";"," ").replace(","," ").replace("  "," ")
            subject_start = i[4]
            subject_end = i[5]
            E_value = i[6]
            identity = i[7]
    
            # save the annotations with the subject always on the '+' strand
            if subject_start > subject_end:
                tmp = subject_start
                subject_start = subject_end
                subject_end = tmp
                tmp = query_start
                query_start = query_end
                query_end = tmp
    
            matchStart = min(query_start, query_end)
            matchEnd = max(query_start, query_end)
            strand = self._getStrand(query_start, query_end)
    
            if dPathID2Matchs.has_key(pathID):
                dPathID2Matchs[pathID].append([matchStart, matchEnd, strand, subject_name, subject_start, subject_end, E_value, identity])
    
            else:
                dPathID2Matchs[pathID] = [[matchStart, matchEnd, strand, subject_name, subject_start, subject_end, E_value, identity]]
    
        return dPathID2Matchs
    
    ## For a specific path ID, organize match data according to the GFF3 format.
    #
    # @param pathID string path ID
    # @param lMatches match list
    # @param seqName string name of the source feature
    # @param source string 'source' field for GFF3 format
    # @param frame string 'frame' field for GFF3 format
    # @return lines string to write in the GFF3 file
    #
    def _organizeEachPathFeature(self, pathID, lMatches, seqName, source, frame):
        lines = ""
        minStart = lMatches[0][0]
        maxEnd = lMatches[0][1]
        minStartSubject = lMatches[0][4]
        maxEndSubject = lMatches[0][5]
        strand = lMatches[0][2]
    
        # for each match
        for i in lMatches:
            if i[0] < minStart:
                minStart = i[0]
            if i[1] > maxEnd:
                maxEnd = i[1]
            if i[4] < minStartSubject:
                minStartSubject = i[4]
            if i[5] > maxEndSubject:
                maxEndSubject = i[5]
    
        target = lMatches[0][3]
        attributes = "ID=ms%s_%s_%s" % (pathID, seqName, target)
        if self._isChado:
            attributes += ";Target=%s+%s+%s" % (target, minStartSubject, maxEndSubject)
        else:
            attributes += ";Target=%s %s %s" % (target, minStartSubject, maxEndSubject)
        if len(lMatches) == 1:
            attributes += ";Identity=%s" % lMatches[0][7]

        lines += "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (seqName, source, "match", minStart, maxEnd, "0.0", strand, frame, attributes)
    
        if len(lMatches) > 1:
            count = 1
            for i in lMatches:
                attributes = "ID=mp%s-%i_%s_%s" % (pathID, count, seqName, target)
                attributes += ";Parent=ms%s%s%s%s%s" % (pathID, "_", seqName, "_", target)
                if self._isChado:
                    attributes += ";Target=%s+%s+%s" % (target, i[4], i[5])
                else:
                    attributes += ";Target=%s %s %s" % (target, i[4], i[5])
                attributes += ";Identity=%s" % i[7]
                lines += "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (seqName, source, "match_part", i[0], i[1], i[6], i[2], frame, attributes)
                count += 1
    
        return lines
    
    ## Save the annotations from the table corresponding to the given sequence.
    #
    # @param key string sequence name
    # @param table string name of the 'set' table
    # @param tierName string name of the tier in the GFF3 file
    #
    def _saveAnnotSet(self, key, table, tierName):
        if self._verbose > 1:
            print "processing '%s' for tier '%s'..." % (key, tierName)
            sys.stdout.flush()
    
        shutil.move("%s.gff3" % key, "%s.gff3_tmp" % key)
        tmpFile = open("%s.gff3_tmp" % key, "r")
        line = tmpFile.readline()
        outFile = open("%s.gff3" % key, "w")
        while True:
            if line == "":
                break
            if line != "##FASTA\n":
                outFile.write(line)
                line = tmpFile.readline()
            else:
                source = tierName
                feature = "region"
                frame = "."
                setString = self._getSetFeatures(table, key, source, feature, frame)
                outFile.write(setString)
                break
        if line == "##FASTA\n":
            while True:
                if line == "":
                    break
                outFile.write(line)
                line = tmpFile.readline()
        tmpFile.close()
        outFile.close()
        os.remove("%s.gff3_tmp" % key)
    
    ## Retrieve the features to write in the GFF3 file.
    #
    # @param table string name of the table recording the annotations (i.e. the features)
    # @param key string name of the sequence (the source feature) on which we want to visualize the matches (the features)
    # @param source string the program that generated the feature (i.e. REPET)
    # @param frame string "." by default (or "+", "-")
    # @return setString string which will be printed in set file 
    # 
    def _getSetFeatures(self, table, key, source, feature, frame):
        setString = ""
        # retrieve all the data about the matches
        qry = "SELECT DISTINCT path,name,start,end FROM %s WHERE chr=\"%s\"" % (table, key)
        self._iDB.execute(qry)
        data = self._iDB.fetchall()
        # organise them into 'match' and 'match_part'
        dPathID2Data = self._gatherSameSetFeatures(data)
        # build the output string
        for pathID in dPathID2Data:
            setString += self._organizeEachSetFeature(pathID, dPathID2Data[pathID], key, source, frame)
        return setString
    
    ## Gather matches with the same path ID.
    #
    # @param data list of string lists results of a SQL request
    # @return dSetID2Matchs dict whose keys are set IDs and values are matches data
    #
    def _gatherSameSetFeatures(self, data):
        dSetID2Matchs = {}
    
        for i in data:
            setID = i[0]
            name = i[1]
            start = i[2]
            end = i[3]
    
            matchStart = min(start, end)
            matchEnd = max(start, end)
            strand = self._getStrand(start, end)
    
            if dSetID2Matchs.has_key(setID):
                dSetID2Matchs[setID].append([name, matchStart, matchEnd, strand])
    
            else:
                dSetID2Matchs[setID] = [[name, matchStart, matchEnd, strand]]
    
        return dSetID2Matchs
    
    ## For a specific set ID, organize match data according to the GFF3 format.
    #
    # @param setID string path ID
    # @param lMatches match list
    # @param seqName string name of the source feature
    # @param source string 'source' field for GFF3 format
    # @param frame string 'frame' field for GFF3 format
    # @return lines string to write in the GFF3 file
    #
    def _organizeEachSetFeature(self, setID, lMatches, seqName, source, frame):
        lines = ""
        minStart = min(lMatches[0][1], lMatches[0][2])
        maxEnd = max(lMatches[0][1], lMatches[0][2])
        strand = lMatches[0][3]
    
        # for each match
        for i in lMatches:
            start = min(i[1],i[2])
            if start < minStart:
                minStart = start
            end = max(i[1],i[2])
            if end > maxEnd:
                maxEnd = end
    
        target = lMatches[0][0].replace("(","").replace(")","").replace("#","")
    
        attributes = "ID=ms%s_%s_%s" % (setID, seqName, target)
        if self._isChado:
            attributes += ";Target=%s+%s+%s" % (target, "1", abs(minStart-maxEnd)+1)
        else:
            attributes += ";Target=%s %s %s" % (target, "1", abs(minStart-maxEnd)+1)
        lines += "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (seqName, source, "match", minStart, maxEnd, "0.0", strand, frame, attributes)
    
        if len(lMatches) > 1:
            count = 1
            for i in lMatches:
                attributes = "ID=mp%s-%i_%s_%s" % (setID, count, seqName, target)
                attributes += ";Parent=ms%s%s%s%s%s" % (setID, "_", seqName, "_", target)
                if self._isChado:
                    attributes += ";Target=%s+%s+%s" % (target, "1", abs(i[1]-i[2])+1)
                else:
                    attributes += ";Target=%s %s %s" % (target, "1", abs(i[1]-i[2])+1)
                lines += "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (seqName, source, "match_part", i[1], i[2], "0.0", i[3], frame, attributes)
                count += 1
    
        return lines
    
    ## Return the strand ('+' if start < end, '-' otherwise).
    #
    # @param start integer start coordinate
    # @param end integer end coordinate
    # @return strand string "+" or "-"
    #
    def _getStrand(self, start, end):
        if start < end:
            return "+"
        else:
            return "-"
        
    def run(self):
        self.checkOptions()
        if self._verbose > 0:
            print "START GFF3Maker"
            sys.stdout.flush()
            
        if self._configFileName == "":
            self._iDB = DbFactory.createInstance()
        else:
            self._iDB = DbFactory.createInstance(self._configFileName)
            
        # create all the ".gff3" files (option '-f')
        if self._inFastaName != "":
            self._createGff3Files()
    
        # save the annotations in the gff3 file
        if self._tablesFileName != "":
            if os.path.exists(self._patternGFF3):
                lGff3Files = [self._patternGFF3]
            else:
                lGff3Files = glob.glob(self._patternGFF3)
                if len(lGff3Files) == 0:
                    print "ERROR: '%s' is neither a file nor a pattern" % self._patternGFF3
                    sys.exit(1)
            for gff3File in lGff3Files:
                if self._verbose > 0:
                    print "handle file '%s'" % gff3File
                    sys.stdout.flush()
                key = gff3File.split(".gff3")[0]
                tablesFile = open(self._tablesFileName, "r")
                lines = tablesFile.readlines()
                for l in lines:
                    if l[0] == "#":
                        continue
                    tok = l.split()
                    if len(tok) == 0:
                        break
                    tierName = tok[0]
                    format = tok[1]
                    table = tok[2]
                    if self._verbose > 0:
                        print "save annotations from table '%s' (format %s)" % (table, format)
                        sys.stdout.flush()
                    if format == "path":
                        self._saveAnnotPath(key, table, tierName)
                    elif format == "set":
                        self._saveAnnotSet(key, table, tierName)
                tablesFile.close()
        self._iDB.close()
    
        if self._verbose > 0:
            print "END GFF3Maker"
            sys.stdout.flush()
            
if __name__ == "__main__":
    iGFF3Maker = GFF3Maker()
    iGFF3Maker.setAttributesFromCmdLine()
    iGFF3Maker.run()
    