#! /usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import sys
import os
import optparse
from pyRepetUnit.commons.checker.CheckerUtils import CheckerUtils
from pyRepetUnit.commons.utils.FileUtils import FileUtils


class LaunchLTRharvest(object):

    def __init__(self, inputFileName = "" , indexname = "", output = "", minlenltr = 100, maxlenltr = 1000, mindistltr = 1000, maxdistltr = 15000, similar = 85, overlaps = "best"):
        self._options = None
        self._inputFileName = inputFileName
        self._indexname = indexname
        self._output = output
        self._minlenltr = minlenltr
        self._maxlenltr = maxlenltr
        self._mindistltr = mindistltr
        self._maxdistltr = maxdistltr
        self._similar = similar
        self._overlaps = overlaps
        
    def setAttributesFromCmdLine(self):
        description = "LaunchLTRharvest can run the LTRharvest program from Genome Tools package."
        
        parser = optparse.OptionParser(description = description)
        parser.add_option('--db', "--input", dest="inputFileName", action="store", type="string", help="input file (fasta type)")
        parser.add_option("--index", "--indexname", dest="indexname", action="store", type="string", default="", help="the indexName, default = inputFileName_index")
        parser.add_option("--out", "--output", dest="output", action="store", type="string", default="", help="output file (fasta type, default = inputFileName_LTRharvest.fa)")
        parser.add_option("--minlen", "--minlenltr",   dest="minlenltr", action="store", type="int", default=100, help="min length of each LTR")
        parser.add_option("--maxlen", "--maxlenltr",   dest="maxlenltr", action="store", type="int", default=1000, help="max length of each LTR")  
        parser.add_option("--mindist", "--mindistltr",   dest="mindistltr", action="store", type="int", default=1000, help="min distance of LTR starting positions")
        parser.add_option("--maxdist", "--maxdistltr",   dest="maxdistltr", action="store", type="int", default=15000, help="max distance of LTR starting positions")
        parser.add_option("--sim", "--similar",   dest="similar", action="store", type="float", default=85, help="min similarity value between two LTRs")
        parser.add_option("--over", "--overlaps",   dest="overlaps", action="store", type="string", default="best", help="no: no overlapping predictions / best: best prediction if an overlap occurs / all: all overlapping predictions (default: best)")
        (self._options, args) = parser.parse_args()
        self._setAttributesFromOptions(self._options)

    def _setAttributesFromOptions(self, options):
        self.setInputFileName(self._options.inputFileName)
        self.setIndexname(self._options.indexname)
        self.setOutput(self._options.output)
        self.setMinlenltr(self._options.minlenltr)
        self.setMaxlenltr(self._options.maxlenltr)
        self.setMindistltr(self._options.mindistltr)
        self.setMaxdistltr(self._options.maxdistltr)
        self.setSimilar(self._options.similar)
        self.setOverlaps(self._options.overlaps)
        
    def setInputFileName(self, inputFileName):
        self._inputFileName = inputFileName
        
    def setIndexname(self, indexname):
        if indexname == "":
            self._indexname = "%s_index" % os.path.splitext(self._inputFileName)[0]
        else:
            self._indexname = indexname
        
    def setOutput(self, output):
        if output == "":
            self._output = "%s_LTRharvest.fa" % os.path.splitext(self._inputFileName)[0]
        else:
            self._output = output
        
    def setMinlenltr(self, minlenltr):
        self._minlenltr = minlenltr
        
    def setMaxlenltr(self, maxlenltr):
        self._maxlenltr = maxlenltr
        
    def setMindistltr(self, mindistltr):
        self._mindistltr = mindistltr
        
    def setMaxdistltr(self, maxdistltr):
        self._maxdistltr = maxdistltr
        
    def setSimilar(self, similar):
        self._similar = similar
        
    def setOverlaps(self, overlaps):
        self._overlaps = overlaps
            
    def checkOptions(self):       
        if self._inputFileName != "":
            if not FileUtils.isRessourceExists(self._inputFileName):
                raise Exception("ERROR: Input file does not exist!")
        else:
            raise Exception("ERROR: No specified --db option!")
            
        if self._indexname == "":
            raise Exception("ERROR: No specified --index option!")
        
        if self._output == "":
            raise Exception("ERROR: Option --out missing!")
         
        if self._minlenltr == "":
            raise Exception("ERROR: No value specified for --minlenltr option!")
        
        if self._maxlenltr == "":
            raise Exception("ERROR: No value specified for --maxlenltr option!")
         
        if self._maxlenltr < self._minlenltr :
            raise Exception("ERROR: the maxlenltr value must be greater than the minlenltr one")
        
        if self._mindistltr == "":
            raise Exception("ERROR: No value specified for --mindistltr option!")
        
        if self._maxdistltr == "":
            raise Exception("ERROR: No value specified for --maxdistltr option!")
        
        if self._maxdistltr < self._mindistltr :
            raise Exception("ERROR: the maxdistltr value must be greater than the mindistltr one")
        
        if self._similar != "":        
            if self._similar > 100.0 or self._similar < 0.0:
                raise Exception("ERROR: the similarity must be between 0-100!")         
        else:
            raise Exception("ERROR: No value specified for --similar option!")
        
        if self._overlaps not in ["no", "best", "all"]:
            raise Exception("ERROR: --overlaps option not in {no, best, all}!")
                                
    def run(self):
        if not CheckerUtils.isExecutableInUserPath("gt") :
            print "ERROR: gt (Genome Tools) must be in your path"
            sys.exit(1)
        else:
            self.checkOptions()
            cmd = 'gt suffixerator -db %s -indexname %s -tis -suf -lcp -des -ssp -sds -dna' % (self._inputFileName, self._indexname)
            os.system(cmd)
            cmd = 'gt ltrharvest -index %s -out %s -minlenltr %s -maxlenltr %s -mindistltr %s -maxdistltr %s -similar %s -overlaps %s' % (self._indexname, self._output, self._minlenltr, self._maxlenltr, self._mindistltr, self._maxdistltr, self._similar, self._overlaps)
            os.system(cmd)
            
if __name__ == "__main__":
    iLaunchLTRharvest = LaunchLTRharvest()
    iLaunchLTRharvest.setAttributesFromCmdLine()
    iLaunchLTRharvest.run()