#!/usr/bin/env python

from PASTEC.SuperAgent import SuperAgent
from PASTEC.Environment import Environment
from pyRepetUnit.profilesDB.ProfilesDB2Map import ProfilesDB2Map
from pyRepetUnit.commons.sql.DbFactory import DbFactory
from pyRepetUnit.commons.utils.RepetOptionParser import RepetOptionParser
from pyRepetUnit.commons.utils.FileUtils import FileUtils
import ConfigParser
import sys
import os

#TODO: check env var if launch as tool (without TEdenovo)
class LaunchPASTEC(object):
    
    def __init__(self, configFileName = "", step = "12", inputFileName = "", verbose = 0, projectName = ""):
        self._configFileName = configFileName
        self._projectName = projectName
        self._step = step
        self._inFileName = inputFileName
        self._verbose = verbose
        
    def setConfigFileName(self, configFileName):
        self._configFileName = configFileName
        
    def setProjectName(self, projectName):
        self._projectName = projectName
        
    def setStep(self, step):
        self._step = step
        
    def setInputFileName(self, inputFileName):
        self._inFileName = inputFileName
        
    def setVerbosity(self, verbose):
        self._verbose = verbose
    
    def setAttributesFromCmdLine(self):
        description = "Insert banks in DB and launch PASTEC\n"
        description += "step 1 : insert profiles, repbaseNT and repbaseAA banks in DB\n"
        description += "step 2 : launch PASTEC\n"
        examples = "EXAMPLE TODO"
        parser = RepetOptionParser(description = description, version = "v1.0", epilog = examples)
        parser.add_option( "-C", "--config", dest = "configFileName", type="string", default = "", help = "configuration file (mandatory for step 1)")
        parser.add_option( "-P", "--project", dest = "projectName", type="string", default = "", help = "project name (if no configuration file or if different")
        parser.add_option( "-S", "--step", dest = "step", action="store", type="string", default = "12", help = "step (default = 12, 1, 2)")
        parser.add_option( "-i", "--input", dest = "inputFileName", type="string", default = "", help = "input file name (fasta)")
        parser.add_option( "-v", "--verbose", dest = "verbose", type="int", default = 1, help = "verbose (default = 0, 1)")
        options, args = parser.parse_args()
        self._setAttributesFromOptions(options)

    def _setAttributesFromOptions(self, options):
        self.setConfigFileName(options.configFileName)
        self.setProjectName(options.projectName)
        self.setStep(options.step)
        self.setInputFileName(options.inputFileName)
        self.setVerbosity(options.verbose)

    def _checkAttributes(self):
        if self._configFileName == "" and self._projectName == "":
            raise Exception("ERROR: No configuration file name or no project name specified in option!")
        if self._configFileName == "" and "1" in self._step:
            raise Exception("ERROR: No configuration file name specified in option!")
        if self._configFileName != "" and not FileUtils.isRessourceExists(self._configFileName):
            raise Exception("ERROR: Can't open configuration file : %s" % self._configFileName)
        if "2" in self._step:
            if self._inFileName == "":
                raise Exception("ERROR: No fasta file specified in option!")
            elif not FileUtils.isRessourceExists(self._inFileName):
                raise Exception("ERROR: Can't open fasta file : %s" % self._inFileName)

    def _createMapFile(self, profilesBank):
        tmpMapFile = "outputProfilesBank.map"
        profilesDB2Map = ProfilesDB2Map()
        profilesDB2Map.setInputFile(profilesBank)
        profilesDB2Map.setOutputFile(tmpMapFile)
        profilesDB2Map.run()
        return tmpMapFile
    
    def run(self):
        #TODO: refactoring + Check if table exists
        self._checkAttributes()
        if self._verbose > 0:
            sys.stdout.write("START LaunchPASTEC\n")
            sys.stdout.flush()
        
        config = None
        if self._configFileName != "":
            self._iDb = DbFactory.createInstance(self._configFileName)
            config = ConfigParser.ConfigParser()
            config.readfp(open(self._configFileName))
        else:
            self._iDb = DbFactory.createInstance()
        if self._projectName == "":
            self._projectName = config.get("project", "project_name")
            
        if '1' in self._step:
            if self._verbose > 0:
                sys.stdout.write("step 1 :\n")
                sys.stdout.flush()
            projectDir = config.get("project", "project_dir")
            sectionName = "detect_features"
                
            isHmmerUsed = config.get(sectionName, "TE_HMMER")
            if isHmmerUsed == "yes":
                profilesBank = "%s/%s" % (projectDir, config.get(sectionName, "TE_HMM_profiles"))
                tableName = "%s_ProfilesDB_map" % self._projectName
                if not self._iDb.doesTableExist(tableName):
                    if not FileUtils.isRessourceExists(profilesBank):
                        raise Exception("ERROR: can't find file '%s'\n" % profilesBank)
                    if self._verbose > 0:
                        msg = "Insertion profiles bank in DB:\n"
                        msg += "\tprofiles bank: %s\n" % profilesBank
                        sys.stdout.write(msg)
                        sys.stdout.flush()
                    tmpMapFile = self._createMapFile(profilesBank)
                    self._iDb.createTable(tableName, "map", tmpMapFile)
                    os.remove(tmpMapFile)
                else:
                    msg = "WARNING: %s already exists. Check it.\n" % tableName
                    sys.stdout.write(msg)
                    sys.stdout.flush()
                
            isBLRtxUsed = config.get(sectionName, "TE_BLRtx")
            isBLRnUsed = config.get(sectionName, "TE_BLRn")
            if isBLRtxUsed == "yes" or isBLRnUsed == "yes":
                repbaseNTBank = "%s/%s" % (projectDir, config.get(sectionName, "TE_nucl_bank"))
                #TODO: change "seq" to "length"
                tableName = "%s_RepbaseNT_seq" % self._projectName
                if not self._iDb.doesTableExist(tableName):
                    if not FileUtils.isRessourceExists(repbaseNTBank):
                        raise Exception("ERROR: can't find file '%s'\n" % repbaseNTBank)
                    if self._verbose > 0:
                        msg = "Insertion repbaseNT bank in DB:\n"
                        msg += "\trepbaseNT bank: %s\n" % repbaseNTBank
                        sys.stdout.write(msg)
                        sys.stdout.flush()
                    self._iDb.createTable(tableName, "length", repbaseNTBank)
                else:
                    msg = "WARNING: %s already exists. Check it.\n" % tableName
                    sys.stdout.write(msg)
                    sys.stdout.flush()
                
            isBLRxUsed = config.get(sectionName, "TE_BLRx")
            if isBLRxUsed == "yes":
                repbaseAABank = "%s/%s" % (projectDir, config.get(sectionName, "TE_prot_bank"))
                #TODO: change "seq" to "length"
                tableName = "%s_RepbaseAA_seq" % self._projectName
                if not self._iDb.doesTableExist(tableName):
                    if not FileUtils.isRessourceExists(repbaseAABank):
                        raise Exception("ERROR: can't find file '%s'\n" % repbaseAABank)
                    if self._verbose > 0:
                        msg = "Insertion repbaseAA bank in DB:\n"
                        msg += "\trepbaseAA bank: %s\n" % repbaseAABank
                        sys.stdout.write(msg)
                        sys.stdout.flush()
                    self._iDb.createTable(tableName, "length", repbaseAABank)
                else:
                    msg = "WARNING: %s already exists. Check it.\n" % tableName
                    sys.stdout.write(msg)
                    sys.stdout.flush()
                
        if '2' in self._step:
            outFileName = "%s.classif" % self._projectName
            if self._verbose > 0:
                msg = "step 2 :\n"
                msg += "input file: %s\n" % self._inFileName
                msg += "output file: %s\n" % outFileName
                sys.stdout.write(msg)
                sys.stdout.flush()
            
            iEnvironment = Environment(self._inFileName, self._iDb, self._projectName, config, self._verbose - 1)
            if config != None:
                iEnvironment.setAttributesFromConfig()
            iSuperAgent = SuperAgent(iEnvironment)
            iSuperAgent.createNeededAgentInstance()
            iSuperAgent.live()
        else:
            self._iDb.close()
            
        if self._verbose > 0:
            msg = "END LaunchPASTEC\n"
            sys.stdout.write(msg)
            sys.stdout.flush()
    
if __name__ == '__main__':
    iLaunchPASTEC = LaunchPASTEC()
    iLaunchPASTEC.setAttributesFromCmdLine()
    iLaunchPASTEC.run()